import multiprocessing as mp
import os
import pathlib
import sys
import threading
import traceback
import warnings

import cloudpickle
import psutil

from . import utils


class Context:
    def __init__(self):
        self.initfns = []
        self.resolver = None
        self.errfile = None
        self.interval = 20
        self.clientkw = {}
        self.serverkw = {}
        self.printlock = threading.Lock()
        self.done = threading.Event()
        self.watcher = None

    @property
    def mp(self):
        return mp.get_context('spawn')

    def options(self):
        return {
            'resolver': self.resolver and cloudpickle.dumps(self.resolver),
            'errfile': self.errfile,
            'interval': self.interval,
            'initfns': self.initfns,
            'clientkw': self.clientkw,
            'serverkw': self.serverkw,
        }

    def setup(
        self,
        resolver=None,
        errfile=None,
        interval=None,
        initfns=None,
        clientkw=None,
        serverkw=None,
        host=None,
        ipv6=None,
    ):
        if resolver:
            if isinstance(resolver, bytes):
                resolver = cloudpickle.loads(resolver)
            assert callable(resolver)
            self.resolver = resolver

        if errfile:
            if isinstance(errfile, str):
                errfile = pathlib.Path(errfile)
            assert hasattr(errfile, 'exists') and hasattr(
                errfile, 'write_text'
            )
            self.errfile = errfile
            self._check_errfile()
            self._install_excepthook()

        if interval:
            assert isinstance(interval, (int, float))
            self.interval = interval

        if initfns:
            self.initfns = []
            for fn in initfns:
                self.initfn(fn, call_now=True)

        if clientkw is not None:
            assert isinstance(clientkw, dict)
            self.clientkw = clientkw

        if serverkw is not None:
            assert isinstance(serverkw, dict)
            self.serverkw = serverkw

        if host is not None:
            self.serverkw['host'] = host

        if ipv6 is not None:
            self.clientkw['ipv6'] = ipv6
            self.serverkw['ipv6'] = ipv6

        if self.errfile and not self.watcher:
            self.watcher = threading.Thread(target=self._watcher, daemon=True)
            self.watcher.start()

    def initfn(self, fn, call_now=True):
        if isinstance(fn, bytes):
            pkl, fn = fn, cloudpickle.loads(fn)
        else:
            pkl, fn = cloudpickle.dumps(fn), fn
        self.initfns.append(pkl)
        call_now and fn()

    def print(self, name, *args, color=None):
        assert args
        if color:
            style = utils.style(color=color)
            reset = utils.style(reset=True)
        else:
            style, reset = '', ''
        with self.printlock:
            print(style + f'[{name}]' + reset, *args)

    def error(self, e, name=None):
        typ, tb = type(e), e.__traceback__
        summary = list(traceback.format_exception_only(typ, e))[0].strip('\n')
        long = ''.join(traceback.format_exception(typ, e, tb)).strip('\n')
        message = f"Error in '{name}' ({summary}):\n{long}"
        with self.printlock:
            style = utils.style(color='red')
            reset = utils.style(reset=True)
            print(style + '\n---\n' + message + reset, file=sys.stderr)
        if self.errfile:
            self.errfile.write_text(message)
            print(f'Wrote errorfile: {self.errfile}', file=sys.stderr)

    def shutdown(self, exitcode):
        children = list(psutil.Process(os.getpid()).children(recursive=True))
        utils.kill_proc(children, timeout=1)
        # TODO
        # if exitcode == 0:
        #   for child in self.children(threading.main_thread()):
        #     child.kill()
        #   os._exit(0)
        # else:
        os._exit(exitcode)

    def close(self):
        self.done.set()
        if self.watcher:
            self.watcher.join()

    def add_worker(self, worker):
        assert hasattr(worker, 'kill')
        current = threading.current_thread()
        if current == threading.main_thread():
            return
        if hasattr(worker, 'thread'):
            assert current != worker.thread
        try:
            current.children.append(worker)
        except AttributeError:
            warnings.warn(
                'Using Portal from plain Python threads is discouraged because '
                + 'they can cause hangs during shutdown.'
            )

    def children(self, thread):
        current = thread or threading.current_thread()
        if current == threading.main_thread():
            return []
        return current.children

    def _watcher(self):
        while not self.done.wait(self.interval):
            self._check_errfile()

    def _install_excepthook(self):
        existing = sys.excepthook

        def patched(typ, val, tb):
            if self.errfile:
                message = ''.join(
                    traceback.format_exception(typ, val, tb)
                ).strip('\n')
                self.errfile.write_text(message)
                print(f'Wrote errorfile: {self.errfile}', file=sys.stderr)
            return existing(typ, val, tb)

        sys.excepthook = patched

    def _check_errfile(self):
        if self.errfile and self.errfile.exists():
            message = f'Shutting down due to error file: {self.errfile}'
            print(message, file=sys.stderr)
            self.shutdown(exitcode=2)


context = Context()


def initfn(fn, call_now=True):
    context.initfn(fn, call_now)


def setup(**kwargs):
    context.setup(**kwargs)


def reset():
    global context
    context.close()
    context = Context()



import collections
import functools
import itertools
import threading
import time
import weakref

from . import client_socket
from . import futures
from . import packlib


class Client:
    def __init__(self, addr, name='Client', maxinflight=16, **kwargs):
        assert 1 <= maxinflight, maxinflight
        self.maxinflight = maxinflight
        self.reqnum = iter(itertools.count(0))
        self.futures = {}
        self.errors = collections.deque()
        self.sendrate = [0, time.time()]
        self.recvrate = [0, time.time()]
        self.waitmean = [0, 0]
        self.cond = threading.Condition()
        self.lock = threading.Lock()
        # Socket is created after the above attributes because the callbacks access
        # some of the attributes.
        self.socket = client_socket.ClientSocket(
            addr, name, start=False, **kwargs
        )
        self.socket.callbacks_recv.append(self._recv)
        self.socket.callbacks_disc.append(self._disc)
        self.socket.callbacks_conn.append(self._conn)
        self.socket.start()

    @property
    def connected(self):
        return self.socket.connected

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError(name)
        try:
            return functools.partial(self.call, name)
        except AttributeError:
            raise ValueError(name)

    def stats(self):
        now = time.time()
        stats = {
            'inflight': len(self.futures),
            'numsend': self.sendrate[0],
            'numrecv': self.recvrate[0],
            'sendrate': self.sendrate[0] / (now - self.sendrate[1]),
            'recvrate': self.recvrate[0] / (now - self.recvrate[1]),
            'waitmean': self.waitmean[0]
            and (self.waitmean[1] / self.waitmean[0]),
        }
        self.sendrate = [0, now]
        self.recvrate = [0, now]
        self.waitmean = [0, 0]
        return stats

    def connect(self, timeout=None):
        return self.socket.connect(timeout)

    def call(self, method, *data):
        reqnum = next(self.reqnum).to_bytes(8, 'little', signed=False)
        start = time.time()
        while len(self.futures) >= self.maxinflight:
            with self.cond:
                self.cond.wait(timeout=0.2)
            try:
                self.socket.require_connection(timeout=0)
            except TimeoutError:
                pass
        with self.lock:
            self.waitmean[1] += time.time() - start
            self.waitmean[0] += 1
            self.sendrate[0] += 1
        if self.errors:  # Raise errors of dropped futures.
            raise self.errors.popleft()
        name = method.encode('utf-8')
        strlen = len(name).to_bytes(8, 'little', signed=False)
        sendargs = (reqnum, strlen, name, *packlib.pack(data))
        future = futures.Future()
        future.sendargs = sendargs
        self.futures[reqnum] = future
        # Store future before sending request because the response may come fast
        # and the response handler runs in the socket's background thread.
        try:
            self.socket.send(*sendargs)
        except client_socket.Disconnected:
            future = self.futures.pop(reqnum)
            future.rai[0] = True
            raise
        return future

    def close(self, timeout=None):
        for future in self.futures.values():
            self._seterr(future, client_socket.Disconnected)
        self.futures.clear()
        self.socket.close(timeout)

    def _recv(self, data):
        assert len(data) >= 16, 'Unexpectedly short response'
        reqnum = bytes(data[:8])
        status = int.from_bytes(data[8:16], 'little', signed=False)
        future = self.futures.pop(reqnum, None)
        if not future:
            existing = sorted(self.futures.keys())
            print(f'Unexpected request number: {reqnum}', existing)
        elif status == 0:
            data = packlib.unpack(data[16:])
            future.set_result(data)
            with self.cond:
                self.cond.notify_all()
        else:
            message = bytes(data[16:]).decode('utf-8')
            self._seterr(future, RuntimeError(message))
            with self.cond:
                self.cond.notify_all()
        try:
            self.socket.recv()
        except AssertionError:
            pass  # Socket is already closed.

    def _disc(self):
        if self.socket.options.autoconn:
            for future in list(self.futures.values()):
                future.resend = True
        else:
            for future in list(self.futures.values()):
                self._seterr(future, client_socket.Disconnected)
            self.futures.clear()

    def _conn(self):
        if self.socket.options.autoconn:
            for future in list(self.futures.values()):
                if getattr(future, 'resend', False):
                    self.socket.send(*future.sendargs)

    def _seterr(self, future, e):
        future.set_error(e)
        rai = future.rai
        weakref.finalize(
            future, lambda: (None if rai[0] else self.errors.append(e))
        )



import threading
import traceback

from . import contextlib
from . import utils


class Thread:
    """
    This thread implementation extends the standard Python thread as follows:

    1. It provide a kill() method that raises a SystemExit error inside the
    thread. This allows killing the thread from the outside.

    2. It keeps track of its subthreads and subprocesses and kills them on
    error to prevent hangs.

    3. The thread is marked as daemon, so that when the parent process dies the
    thread is killed, preventing hangs.
    """

    def __init__(self, fn, *args, name=None, start=False):
        self.fn = fn
        self.excode = None
        name = name or getattr(fn, '__name__', 'thread')
        self.thread = threading.Thread(
            target=self._wrapper, args=args, name=name, daemon=True
        )
        self.thread.children = []
        self.started = False
        self.ready = threading.Barrier(2)
        contextlib.context.add_worker(self)
        start and self.start()

    @property
    def name(self):
        return self.thread.name

    @property
    def tid(self):
        return self.thetid

    @property
    def running(self):
        if not self.started:
            return False
        return self.thread.is_alive()

    @property
    def exitcode(self):
        return self.excode

    def start(self):
        assert not self.started
        self.started = True
        self.thread.start()
        self.ready.wait()
        return self

    def join(self, timeout=None):
        self.thread.join(timeout)
        return self

    def kill(self, timeout=1.0):
        assert self.thread != threading.current_thread()
        for child in contextlib.context.children(self.thread):
            child.kill(timeout)
        utils.kill_thread(self.thread, timeout)
        return self

    def __repr__(self):
        attrs = ('name', 'ident', 'running', 'exitcode')
        attrs = [f'{k}={getattr(self, k)}' for k in attrs]
        return 'Thread(' + ', '.join(attrs) + ')'

    def _wrapper(self, *args):
        try:
            self.ready.wait()
            exitcode = self.fn(*args)
            exitcode = exitcode if isinstance(exitcode, int) else 0
            self.excode = exitcode
        except (SystemExit, KeyboardInterrupt) as e:
            compact = traceback.format_tb(e.__traceback__)
            compact = '\n'.join([line.split('\n', 1)[0] for line in compact])
            print(f"Killed thread '{self.name}' at:\n{compact}")
            [x.kill(0.1) for x in contextlib.context.children(self.thread)]
            self.excode = 2
        except Exception as e:
            contextlib.context.error(e, self.name)
            contextlib.context.shutdown(exitcode=1)



import math
import weakref
from multiprocessing import shared_memory

import numpy as np


class SharedArray:
    def __init__(self, shape, dtype, name=None):
        if name:
            self.shm = shared_memory.SharedMemory(name=name)
        else:
            size = math.prod(shape) * np.dtype(dtype).itemsize
            self.shm = shared_memory.SharedMemory(create=True, size=size)
        self.arr = np.ndarray(shape, dtype, self.shm.buf)
        # This unlinks the shared memory buffer, but it will survive until the last
        # process closes up their file pointer. This could cause a problem if a
        # SharedArray is serialize and the Python object goes out of scope before
        # trying to deserialize it (in the same or a different process).
        weakref.finalize(self.arr, self.close)

    @property
    def name(self):
        return self.shm.name

    @property
    def array(self):
        return self.arr

    def result(self):
        return self.arr

    def close(self):
        self.arr = None
        try:
            self.shm.unlink()
        except FileNotFoundError:
            pass

    def __getstate__(self):
        return (self.arr.shape, self.arr.dtype.str, self.shm.name)

    def __setstate__(self, args):
        self.__init__(*args)



import collections
import os
import weakref

import numpy as np


class SendBuffer:
    def __init__(self, *buffers, maxsize=None):
        for buffer in buffers:
            types = (bytes, bytearray, memoryview)
            assert isinstance(buffer, types), type(buffer)
            assert not isinstance(buffer, memoryview) or buffer.c_contiguous
        buffers = tuple(
            x.cast('c') if isinstance(x, memoryview) else x for x in buffers
        )
        length = sum(len(x) for x in buffers)
        assert all(len(x) for x in buffers)
        assert 1 <= length, length
        assert not maxsize or length <= length, (length, maxsize)
        lenbuf = length.to_bytes(8, 'little', signed=False)
        self.buffers = [lenbuf, *buffers]
        self.remaining = collections.deque(self.buffers)
        self.pos = 0

    def __repr__(self):
        lens = [len(x) for x in self.buffers]
        left = [len(x) for x in self.remaining]
        return f'SendBuffer(pos={self.pos}, lengths={lens} remaining={left})'

    def reset(self):
        self.remaining = collections.deque(self.buffers)
        self.pos = 0

    def send(self, sock):
        first, *others = self.remaining
        assert self.pos < len(first)
        # The writev() call blocks but seems to be slightly faster than sendmsg().
        size = os.writev(
            sock.fileno(), [memoryview(first)[self.pos :], *others]
        )
        # size = sock.sendmsg(
        #     [memoryview(first)[self.pos:], *others], (), socket.MSG_DONTWAIT)
        if size == 0:
            raise ConnectionResetError
        assert 0 <= size, size
        self.pos += max(0, size)
        while self.remaining and self.pos >= len(self.remaining[0]):
            self.pos -= len(self.remaining.popleft())
        return size

    def done(self):
        return not self.remaining


class RecvBuffer:
    def __init__(self, maxsize):
        self.maxsize = maxsize
        self.lenbuf = b''
        self.buffer = None
        self.pos = 0

    def __repr__(self):
        length = self.buffer and len(self.buffer)
        return f'RecvBuffer(pos={self.pos}, length={length})'

    def recv(self, sock):
        if self.buffer is None:
            part = sock.recv(8 - len(self.lenbuf))
            self.lenbuf += part
            size = len(part)
            if len(self.lenbuf) == 8:
                length = int.from_bytes(self.lenbuf, 'little', signed=False)
                assert 1 <= length <= self.maxsize, (1, length, self.maxsize)
                # We use Numpy to allocate uninitialized memory because Python's
                # `bytearray(length)` zero initializes which is slow. This also means
                # the buffer cannot be pickled accidentally unless explicitly converted
                # to a `bytes()` object, which is a nice bonus for preventing
                # performance bugs in user code.
                arr = np.empty(length, np.uint8)
                self.buffer = memoryview(arr.data)
                weakref.finalize(self.buffer, lambda arr=arr: arr)
                self.pos = 0
        else:
            size = sock.recv_into(self.buffer[self.pos :])
            self.pos += max(0, size)
            info = (0, self.pos, len(self.buffer))
            assert 0 <= self.pos <= len(self.buffer), info
        if size == 0:
            raise ConnectionResetError
        return size

    def done(self):
        return self.buffer and self.pos == len(self.buffer)

    def result(self):
        return self.buffer



import concurrent.futures


class ThreadPool:
    def __init__(self, workers, name):
        self.pool = concurrent.futures.ThreadPoolExecutor(workers, name)

    def submit(self, fn, *args, **kwargs):
        future = self.pool.submit(fn, *args, **kwargs)
        # Prevent deamon threads from hanging due to exit handlers registered by
        # the concurrent.futures modules.
        concurrent.futures.thread._threads_queues.clear()
        return future

    def close(self, wait=False):
        self.pool.shutdown(wait=wait)



import threading


class Future:
    def __init__(self):
        self.rai = [False]
        self.fns = []
        self.con = threading.Condition()
        self.don = False
        self.res = None
        self.err = None

    def __repr__(self):
        if not self.don:
            info = 'done=False'
        elif self.err:
            info = f"done=True, error='{self.err}' raised={self.rai[0]}"
        else:
            info = 'done=True'
        return f'Future({info})'

    def wait(self, timeout=None):
        if self.don:
            return self.don
        with self.con:
            return self.con.wait(timeout)

    def done(self):
        return self.don

    def error(self):
        return self.err

    def result(self, timeout=None):
        if not self.wait(timeout):
            raise TimeoutError
        assert self.don
        if self.err is None:
            return self.res
        if not self.rai[0]:
            self.rai[0] = True
            raise self.err

    def set_result(self, result):
        with self.con:
            assert not self.don
            self.don = True
            self.res = result
            for fn in self.fns:
                fn(self)
            self.con.notify_all()

    def set_error(self, e):
        with self.con:
            assert not self.don
            self.don = True
            self.err = e
            for fn in self.fns:
                fn(self)
            self.con.notify_all()

    def add_callback(self, fn):
        with self.con:
            self.fns.append(fn)
            if self.don:
                fn(self)

    def remove_callback(self, fn):
        self.fns.remove(fn)


def wait(futures, timeout=None, amount=None):
    if amount is None:
        amount = len(futures)
    waiter = _Waiter(amount)
    for future in futures:
        future.add_callback(waiter)
    result = waiter.wait(timeout)
    for future in futures:
        future.remove_callback(waiter)
    if result:
        return tuple(waiter.completed)
    else:
        return False


class _Waiter:
    def __init__(self, amount):
        self.remaining = amount
        self.event = threading.Event()
        self.lock = threading.Lock()
        self.completed = []

    def wait(self, timeout=None):
        return self.event.wait(timeout)

    def __call__(self, future):
        self.completed.append(future)
        if self.remaining <= 0:
            return
        with self.lock:
            self.remaining -= 1
            if self.remaining <= 0:
                self.event.set()



import atexit
import errno
import traceback

import cloudpickle
import psutil

from . import contextlib
from . import utils


class Process:
    """
    This process implementation extends the standard Python process as follows:

    1. It provides a kill() method that not only terminates the process itself
    but also all its nested child processes. This prevents subprocesses from
    lingering in the background after the Python program has ended.

    2. The process terminates its nested child processes when it encounters an
    error. This prevents lingering subprocesses on error.

    3. When the parent process encounters an error, its subprocess will be
    killed via `atexit`, preventing hangs.

    4. It inherits the context object of its parent process, which provides
    cloudpickled initializer functions for each nested child process and error
    file watching for global shutdown on error.

    5. Standard Python subprocesses do not reliably return the running() state
    of the process. This class makes it more reliable.
    """

    def __init__(self, fn, *args, name=None, start=False):
        name = name or getattr(fn, '__name__', 'process')
        fn = cloudpickle.dumps(fn)
        options = contextlib.context.options()
        self.ready = contextlib.context.mp.Barrier(2)
        self.process = contextlib.context.mp.Process(
            target=self._wrapper,
            name=name,
            args=(options, self.ready, name, fn, args),
        )
        self.started = False
        self.killed = False
        self.thepid = None
        atexit.register(self.kill)
        contextlib.context.add_worker(self)
        start and self.start()

    @property
    def name(self):
        return self.process.name

    @property
    def pid(self):
        return self.thepid

    @property
    def running(self):
        if not self.started:
            return False
        if not self.process.is_alive():
            return False
        return utils.proc_alive(self.pid)

    @property
    def exitcode(self):
        exitcode = self.process.exitcode
        if self.killed and exitcode is None:
            return -9
        return exitcode

    def start(self):
        assert not self.started
        self.started = True
        self.process.start()
        self.ready.wait()
        self.thepid = self.process.pid
        assert self.thepid is not None
        return self

    def join(self, timeout=None):
        assert self.started
        self.process.join(timeout)
        return self

    def kill(self, timeout=1):
        assert self.started
        try:
            children = list(psutil.Process(self.pid).children(recursive=True))
        except psutil.NoSuchProcess:
            children = []
        try:
            self.process.terminate()
            self.process.join(timeout)
            self.process.kill()
            utils.kill_proc(children, timeout)
        except OSError as e:
            if e.errno != errno.ESRCH:
                contextlib.context.error(e, self.name)
                contextlib.context.shutdown(exitcode=1)
        self.killed = True
        return self

    def __repr__(self):
        attrs = ('name', 'pid', 'running', 'exitcode')
        attrs = [f'{k}={getattr(self, k)}' for k in attrs]
        return 'Process(' + ', '.join(attrs) + ')'

    @staticmethod
    def _wrapper(options, ready, name, fn, args):
        exitcode = 0
        try:
            ready.wait()
            contextlib.setup(**options)
            fn = cloudpickle.loads(fn)
            exitcode = fn(*args)
            exitcode = exitcode if isinstance(exitcode, int) else 0
        except (SystemExit, KeyboardInterrupt) as e:
            compact = traceback.format_tb(e.__traceback__)
            compact = '\n'.join([line.split('\n', 1)[0] for line in compact])
            print(f"Killed process '{name}' at:\n{compact}")
            exitcode = 2
        except Exception as e:
            contextlib.context.error(e, name)
            exitcode = 1
        finally:
            contextlib.context.shutdown(exitcode)



import ctypes
import errno
import os
import socket
import sys
import threading
import time

import psutil

from . import contextlib


def run(workers, duration=None):
    [None if x.started else x.start() for x in workers]
    start = time.time()
    while True:
        time.sleep(0.1)
        if duration and time.time() - start >= duration:
            print(f'Shutting down workers after {duration} seconds.')
            [x.kill() for x in workers]
            return
        if all(x.exitcode == 0 for x in workers):
            print('All workers terminated successfully.')
            return
        errored = [x for x in workers if x.exitcode not in (None, 0)]
        if errored:
            time.sleep(0.1)  # Wait for workers to print their error messages.
            name = errored[0].name
            code = errored[0].exitcode
            print(f"Shutting down workers due to crash in '{name}' ({code}).")
            [x.kill() for x in workers]
            raise RuntimeError(f"'{name}' crashed with exit code {code}")


def kill_thread(threads, timeout=1):
    threads = threads if isinstance(threads, (list, tuple)) else [threads]
    for thread in threads:
        if thread.native_id is None:
            # Wait because thread may currently be starting.
            time.sleep(0.2)
        matches = [k for k, v in threading._active.items() if v is thread]
        if not matches:
            continue
        ident = matches[0]
        result = ctypes.pythonapi.PyThreadState_SetAsyncExc(
            ctypes.c_long(ident), ctypes.py_object(SystemExit)
        )
        if result > 1:
            ctypes.pythonapi.PyThreadState_SetAsyncExc(
                ctypes.c_long(ident), None
            )
    start = time.time()
    [x.join(max(0.1, timeout - (time.time() - start))) for x in threads]
    for thread in threads:
        if thread.is_alive():
            print('Killed thread is still alive.')


def kill_proc(procs, timeout=1):
    def eachproc(fn, procs):
        result = []
        for proc in list(procs):
            try:
                result.append(fn(proc))
            except psutil.NoSuchProcess:
                pass
        return result

    # Collect all sub processes.
    procs = procs if isinstance(procs, (list, tuple)) else [procs]
    procs = eachproc(
        lambda p: psutil.Process(p) if isinstance(p, int) else p, procs
    )
    eachproc(lambda p: procs.extend(p.children(recursive=True)), procs)
    procs = list(set(procs))
    # Send SIGINT to attempt graceful shutdown.
    eachproc(lambda p: p.terminate(), procs)
    _, procs = psutil.wait_procs(procs, timeout)
    # Send SIGTERM to remaining processes to force exit.
    eachproc(lambda p: p.kill(), procs)
    # Should never happen but print warning if any survived.
    eachproc(
        lambda p: (
            print('Killed subprocess is still alive.')
            if proc_alive(p.pid)
            else None
        ),
        procs,
    )


def proc_alive(pid):
    try:
        if psutil.Process(pid).status() == psutil.STATUS_ZOMBIE:
            return False
    except psutil.NoSuchProcess:
        return False
    try:
        os.kill(pid, 0)
    except OSError as e:
        if e.errno == errno.ESRCH:
            return False
    return True


def free_port():
    # Return a port that is currently free. This function is not thread or
    # process safe, because there is no way to guarantee that the port will still
    # be free at the time it will be used.
    ipv6 = contextlib.context.serverkw.get('ipv6', False)
    host = contextlib.context.serverkw.get('host', '')
    if ipv6:
        family, addr = socket.AF_INET6, (host or '::', 0, 0, 0)
    else:
        family, addr = socket.AF_INET, (host or '0.0.0.0', 0)
    sock = socket.socket(family, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(addr)
    port = sock.getsockname()[1]
    sock.close()
    return port


def style(color=None, background=None, bold=None, underline=None, reset=None):
    if not sys.stdout.isatty():
        return ''
    escseq = lambda parts: '\033[' + ';'.join(parts) + 'm'
    colors = dict(
        black=0, red=1, green=2, yellow=3, blue=4, magenta=5, cyan=6, white=7
    )
    parts = []
    if reset:
        parts.append(escseq('0'))
    if color or bold or underline:
        args = ['3' + (str(colors[color]) if color else '9')]
        bold and args.append('1')
        underline and args.append('4')
        parts.append(escseq(args))
    if background:
        parts.append(escseq('4' + str(colors[background])))
    return ''.join(parts)



import threading

import numpy as np
import portal

from . import client
from . import packlib
from . import process
from . import server
from . import server_socket
from . import sharray
from . import thread
from . import utils


class BatchServer:
    def __init__(
        self,
        port,
        name='Server',
        workers=1,
        errors=True,
        process=True,
        shmem=False,
        **kwargs,
    ):
        inner_port = utils.free_port()
        while inner_port == port:
            inner_port = utils.free_port()
        self.name = name
        self.server = server.Server(
            inner_port, name, workers, errors, **kwargs
        )
        if process:
            self.running = portal.context.mp.Event()
        else:
            self.running = threading.Event()
        self.process = process
        self.batsizes = {}
        self.batargs = (
            self.running,
            port,
            inner_port,
            f'{name}Batcher',
            self.batsizes,
            errors,
            shmem,
            kwargs,
        )
        self.started = False

    def bind(self, name, workfn, donefn=None, batch=0, workers=0):
        assert not self.started
        self.batsizes[name] = batch
        self.server.bind(name, workfn, donefn, workers=workers)

    def start(self, block=True):
        assert not self.started
        self.started = True
        self.running.set()
        if self.process:
            self.batcher = process.Process(
                batcher, *self.batargs, name=f'{self.name}Batcher', start=True
            )
        else:
            self.batcher = thread.Thread(
                batcher, *self.batargs, name=f'{self.name}Batcher', start=True
            )
        self.server.start(block=block)

    def close(self, timeout=None):
        assert self.started
        self.running.clear()
        self.server.close(timeout)
        self.batcher.join(timeout)
        self.batcher.kill()

    def stats(self):
        return self.server.stats()

    def __enter__(self):
        self.start(block=False)
        return self

    def __exit__(self, *e):
        self.close()


def batcher(
    running, outer_port, inner_port, name, batsizes, errors, shmem, kwargs
):
    def maybe_recv(outer, inner, jobs, batches):
        if not running.is_set():  # Do not accept further requests.
            return
        try:
            addr, data = outer.recv(timeout=0.0001)
        except TimeoutError:
            return
        reqnum = bytes(data[:8])
        data = data[8:]
        strlen = int.from_bytes(data[:8], 'little', signed=False)
        data = data[8:]
        name, data = bytes(data[:strlen]).decode('utf-8'), data[strlen:]
        if name not in batsizes:
            send_error(addr, reqnum, 3, f'Unknown method {name}')
            return
        data = packlib.unpack(data)
        batch_size = batsizes[name]
        if not batch_size:
            job = inner.call(name, *data)
            job.args = (False, addr, reqnum)
            jobs.append(job)
            return
        leaves, structure = packlib.tree_flatten(data)
        leaves = [np.asarray(x) for x in leaves]
        if any(x.dtype == object for x in leaves):
            send_error(addr, reqnum, 5, 'Only array arguments can be batched.')
            return
        if name not in batches:
            if shmem:
                buffers = [
                    sharray.SharedArray((batch_size, *leaf.shape), leaf.dtype)
                    for leaf in leaves
                ]
            else:
                buffers = [
                    np.empty((batch_size, *leaf.shape), leaf.dtype)
                    for leaf in leaves
                ]
            batches[name] = ([], [], structure, buffers)
        addrs, reqnums, reference, buffers = batches[name]
        if structure != reference:
            message = (
                f'Argument structure {structure} does not match previous '
                + f'requests with structure {reference} for batched server '
                + f'method {name}.'
            )
            send_error(addr, reqnum, 6, message)
            return
        index = len(addrs)
        addrs.append(addr)
        reqnums.append(reqnum)
        for buffer, leaf in zip(buffers, leaves):
            buffer[index] = leaf
        if len(addrs) == batch_size:
            del batches[name]
            data = packlib.tree_unflatten(buffers, reference)
            job = inner.call(name, *data)
            job.args = (True, addrs, reqnums)
            jobs.append(job)

    def maybe_send(outer, inner, jobs):
        done, waiting = [], []
        [done.append(x) if x.done() else waiting.append(x) for x in jobs]
        for job in done:
            batched, addr, reqnum = job.args
            try:
                result = job.result()
            except RuntimeError as e:
                if batched:
                    for i, (addr, reqnum) in enumerate(zip(addr, reqnum)):
                        send_error(addr, reqnum, 6, e.args[0])
                else:
                    send_error(addr, reqnum, 6, e.args[0])
                continue
            status = int(0).to_bytes(8, 'little', signed=True)
            if batched:
                for i, (addr, reqnum) in enumerate(zip(addr, reqnum)):
                    data = packlib.tree_map(lambda x: x[i], result)
                    data = packlib.pack(data)
                    outer.send(addr, reqnum, status, *data)
            else:
                data = packlib.pack(result)
                outer.send(addr, reqnum, status, *data)
        return waiting

    def send_error(addr, reqnum, status, message):
        assert 1 <= status, status
        status = status.to_bytes(8, 'little', signed=False)
        data = message.encode('utf-8')
        outer.send(addr, reqnum, status, data)
        if errors:
            raise RuntimeError(message)

    try:
        outer = server_socket.ServerSocket(
            outer_port, f'{name}Server', **kwargs
        )
        inner = client.Client(inner_port, f'{name}Client', **kwargs)
        batches = {}  # {method: ([addr], [reqnum], structure, [array])}
        jobs = []
        shutdown = False
        while running.is_set() or jobs:
            if running.is_set():
                maybe_recv(outer, inner, jobs, batches)
            elif not shutdown:
                shutdown = True
                outer.shutdown()
            jobs = maybe_send(outer, inner, jobs)
    finally:
        outer.close()
        inner.close()
